
------------------------------------------------------------------------------------------
--
-- Copyright (C) 2016 Intel, Inc
-- All Rights Reserved.
--
-- Description: LUA script used to serve as a POC for the eventual production rules
--
-- Author:      apullat
-- Date:        04/11/2016
--
------------------------------------------------------------------------------------------

-- Tag That identifies the script file. This will appear in the logs
TAG = "MessagingRules.lua"


-- ==================================================================
-- Section on Definitions used by the LUA Scripts
-- ==================================================================
NDefines = {

    -- ==================================================================
    -- Creating a master list of all available messaging types in this sheet
    -- ==================================================================
NMessages = {
    -- 0 to 99 belong to operational messages
    MESSAGE_UNKNOWN = 0,                                                -- Unknown message

    -- 100 to 999 belong to feature re-engagement messages
    MESSAGE_REENGAGE_MAIN_SCREEN_AFTER_7_INACTIVE_DAYS = 100,           -- Reintroduce the user to the main sceren
    MESSAGE_REENGAGE_REGISTRATION_AFTER_3_UNREGISTERED_DAYS = 200,      -- Remind the user to register the product

    -- 1000 to 1999 belong to promoting other products
    MESSAGE_PROMOTE_TRUE_KEY_UPON_APP_LAUNCH = 1000,                    -- User has launched a TK compatible app
    MESSAGE_PROMOTE_TRUE_KEY_UPON_URL_NAVIGATION = 1001,                -- User has browsed to a TK compatible web page

},


    -- ==================================================================
    -- All possible events that can initiate a message
    -- ==================================================================

NEvents = {             -- Must match whatever is defined in the Java code
    -- 0 to 10 belong to operational events
    EVENT_UNKNOWN = 0,                          -- Unknown event

    -- 10 to 99 belong to timer/scheduler events
    EVENT_ON_SCHEDULED_TIMER = 10,              -- Vanilla Scheduled Timer

    -- 100 to 199 belong to events fired by other apps
    EVENT_APP_INSTALL = 100,                    -- Third-party app installed
    EVENT_APP_LAUNCH = 101,                     -- Third-party app launched

    -- 200 to 299 belong to web browsing events
    EVENT_ON_URL_NAVIGATION = 200,              -- User has opened a URL on their browser

},

    -- ==================================================================
    -- Cadence and Cooldown variables to prevent overtly spamming the user
    -- ==================================================================

NCadence = {            -- Allows throttling of messaging
    COOLDOWN_PERIOD_INITIAL = 7,                -- Don't annoy users for the first 7 days
    COOLDOWN_PERIOD_PROMOS = 7,                 -- We won't show more than one promo every 7 days
    COOLDOWN_PERIOD_REMIND_ME_LATER = 7,        -- The standard "remind me later" interval for each message
    COOLDOWN_PERIOD_REENGAGEMENT = 1,           -- We'll not message the user more than once per day

    MAX_COUNT_TRUEKEY_PROMOS = 2,               -- We won't promote True Key more than 2 times
    MIN_COUNT_TRUEKEY_TARGETS = 4,              -- We won't promote True Key unless we identify the user as a valid target

},

    -- ==================================================================
    -- Presentation patterns supported by the client
    -- ==================================================================

NPresentation = {       -- Must match whatever is defined in the Java code
    PRESENTATION_EMPTY                          = 0,    -- No relevant message
    PRESENTATION_SIMPLE_HEADS_UP_NOTIFICATION   = 10,   -- A heads-up high priority notification
    PRESENTATION_SIMPLE_NOTIFICATION            = 11,   -- A simple, non-critical notification

},

    -- ==================================================================
    -- Calls to Action supported by the client
    -- ==================================================================

NCallsToAction = {      -- Must match whatever is defined in the Java code
    CALL_TO_ACTION_DO_NOTHING                   = 0,    -- Do nothing
    CALL_TO_ACTION_OPEN_APP_URL                 = 10,   -- Take the user to the corresponding App Store
    CALL_TO_ACTION_OPEN_WEB_URL                 = 11,   -- Take the user to a web URL
},

}


-- Create global variables for easy use later
recommendation = nil
tracer = nil
event = nil
appContext = nil
msgContext = nil

-- ==================================================================
-- The Evaluate function will re-evaluate all rules.
--
-- Maintaining the ordering of the list of params below is critical!
-- @param ret_val: The messaging recommendation returned after evaluation
-- @param util_tracer: Allows the script to log to logcat
-- @param ctx_event: The Trigger Event responsible for a re-evaluation
-- @param ctx_app: Additional context exposed by the app
-- @param ctx_msg: Context persistance made available to the messaging rules
--
-- ==================================================================
function Evaluate(ret_val, util_tracer, ctx_event, ctx_app, ctx_msg)

    -- Update global variables for easy use later
    recommendation =    ret_val
    tracer =            util_tracer
    event =             ctx_event
    appContext =        ctx_app
    msgContext =        ctx_msg

    local evtType = event:getEventType()
    tracer:Trace(TAG, "Evaluate invoked from LUA Script for Event type: " .. evtType)

    if (evtType == "EVENT_ON_DEVICE_HOME_SCREEN") then
        HandleOnDeviceHomeScreenEvent()
    elseif (evtType == "EVENT_ON_APP_LAUNCH") then
        HandleOnAppLaunchEvent()
    elseif (evtType == "EVENT_ON_URL_NAVIGATE") then
        HandleOnURLNavigateEvent()
    elseif (evtType == "EVENT_ON_MESSAGE_SHOWN") then
        HandleOnMessageShownEvent()
    end

end

function HandleOnDeviceHomeScreenEvent()

    -- The Device Home screen is relevant in the following use cases (prioritized from more to less important)
    -- 1) ** Promotion of True Key courtesy a slider (supported by script) **

    -- We will return as soon as we've picked a promotion - we do not stack multiple promotions

    -- Promote True Key
    if canPromoteTrueKeyOnHomeScreen() then doPromoteTrueKeyOnHomeScreen() return end

end

function HandleOnAppLaunchEvent()

    -- The App Launch event is relevant in the following use cases (prioritized from more to less important)
    -- 1) ** NOT A PROMOTION - If the app is supported by TK, track this state for messaging purposes **

    -- True Key messaging state updates - can cascade event
    if isAppSupportedByTrueKey(event:getPackageName()) then updateTrueKeyMessagingStateForAppLaunch() end

end

function HandleOnURLNavigateEvent()

    -- The URL navigation event is relevant in the following use cases (prioritized from more to less important)
    -- 1) ** NOT A PROMOTION - If the url is supported by TK, track this state for messaging purposes **

    -- True Key messaging state updates - can cascade event
    tracer:Trace(TAG, "HandleOnURLNavigateEvent")
    if isURLSupportedByTrueKey(event:getUrl()) then updateTrueKeyMessagingStateForURLNavigate() end

end

function HandleOnMessageShownEvent()

    -- The Message Shown event allows the rules engine to close the loop
    -- and do state management depending on the message type
    tracer:Trace(TAG, "Message was shown to the user:")
    local messageId = event:getMessageId()
    tracer:Trace(TAG, "Message was shown to the user: " .. messageId)

    -- Persist context on message seen
    msgContext:OnMessageSeen( messageId, appContext:getLong("Device.CurrentTime") )

    if (messageId == "MESSAGE_PROMOTE_TRUE_KEY_ON_HOME_SCREEN") then
        handleOnTrueKeyPromoShownOnHomeScreen()
    end

end


-- ==================================================================
-- Section for utility scripts
-- ==================================================================

-- Calculation in milliseconds
function hasCooldownPeriodExpired (start, finish, interval)
    tracer:Trace(TAG,"start " ..start)
    tracer:Trace(TAG, "finish " ..finish)
    tracer:Trace(TAG, "interval " ..interval)
    tracer:Trace(TAG, "interval in milliseconds " ..(interval * 24 * 60 * 60 * 1000))
    return ((finish - start) > (interval * 24 * 60 * 60 * 1000))
end

-- ==================================================================
-- Section on Scripts used to evaluate and run True Key promotions
-- ==================================================================

-- User has given us an opportunity on the home screen to promote True Key
-- Message name: MESSAGE_PROMOTE_TRUE_KEY_ON_HOME_SCREEN


function canPromoteTrueKeyOnHomeScreen()
    tracer:Trace(TAG, "Verifying TK prerequisites!")
    if ( not canPromoteTrueKeyOnHomeScreen_intransigentChecks() ) then return false end
    if ( not canPromoteTrueKeyOnHomeScreen_contextualChecks() ) then return false end
    if ( not appContext:getBoolean("Device.GreenState")) then return false end

    return true
end


function canPromoteTrueKeyOnHomeScreen_intransigentChecks()
    tracer:Trace(TAG, "canPromoteTrueKeyOnHomeScreen_intransigentChecks")
    -- This function checks for the most "stubborn" conditions - those that almost never vary
    -- Helps rapidly make decisions thereby optimizing script performance
    if ( not appContext:getBoolean("Device.PlayStoreState")) then return false end
    -- Should only be shown for Android OS version 4.0.3 (Ice Cream Sandwich MR1 or 15) and above
    if (appContext:getInt("Device.OS.Version") < 15) then return false end

    -- Does true key already exists on the device? OR
    -- If user installed true key atleast once on device, do not show trigger
    if ( wasTrueKeyEverInstalledOnTheDevice() ) then return false end

    -- True key promotion should be shown max of 2 times
   if (not (msgContext:getMessageSeenCount("MESSAGE_PROMOTE_TRUE_KEY_ON_HOME_SCREEN") < NDefines.NCadence.MAX_COUNT_TRUEKEY_PROMOS)) then return false end

    -- If the user has expressly selected "Do not disturb"/"Never remind again" for a message, then bail out
    if ( msgContext:getMessageDNDStatus("MESSAGE_PROMOTE_TRUE_KEY_ON_HOME_SCREEN") ) then return false end
    tracer:Trace(TAG, "canPromoteTrueKeyOnHomeScreen_intransigentChecks return true")
    return true

end


function canPromoteTrueKeyOnHomeScreen_contextualChecks()
    tracer:Trace(TAG, "canPromoteTrueKeyOnHomeScreen_contextualChecks")
    -- This function checks for the most transient conditions - those that vary often
    -- Involves more complex calculations and can affect outcomes in wide varieties

    -- TEMPORARILY DISABLED: There needs to be a cooldown period of 7 days between each message display
    -- if (not hasCooldownPeriodExpired(msgContext:getLastMessageTimeOverall(), appContext:getLong("Device.CurrentTime"), NDefines.NCadence.COOLDOWN_PERIOD_PROMOS)) then return false end

    -- “Previous” app launched/url launched is supported by true key
    if ( (not hasTrueKeySupportedAppRecentlyBeenOpened()) and (not hasTrueKeySupportedURLRecentlyBeenNavigated()) ) then
        tracer:Trace(TAG, "hasTrueKeySupportedURLRecentlyBeenNavigated return false")
        return false
    end

    -- Verify if user has opened atleast 4 unique true key supported apps/urls after the first true key banner is shown
    if ( not hasSufficientUserVisitsToTrueKeyCandidates() ) then
        tracer:Trace(TAG, "hasSufficientUserVisitsToTrueKeyCandidates return false")
        return false
    end

    -- If the user has asked us to "Remind me later", then check to see if the cooldown period has expired
    if ( not hasCooldownPeriodExpired( msgContext:getRemindMeLaterSelectionTime("MESSAGE_PROMOTE_TRUE_KEY_ON_HOME_SCREEN"),
        appContext:getLong("Device.CurrentTime"), NDefines.NCadence.COOLDOWN_PERIOD_REMIND_ME_LATER ) ) then return false end
    tracer:Trace(TAG, "canPromoteTrueKeyOnHomeScreen_contextualChecks returning true")
    return true

end


function doPromoteTrueKeyOnHomeScreen()
    tracer:Trace(TAG, "Promote True Key on Home Screen!")



    -- The promotion is via updating the recommendation object
    -- The actual promotion will be handled by the Java layer based on the information we send it

    recommendation:setMessageId("MESSAGE_PROMOTE_TRUE_KEY_ON_HOME_SCREEN")
    recommendation:setPresentationType("PRESENTATION_OUT_OF_APP_SIDE_SLIDER")
    recommendation:setUri("MESSAGE_PROMOTE_TRUE_KEY_ON_HOME_SCREEN.json")
    recommendation:setAction("KEY_SLIDER_POSITIVE_BUTTON", NDefines.NCallsToAction.CALL_TO_ACTION_OPEN_APP_URL)
    recommendation:setAction("KEY_SLIDER_NEGATIVE_BUTTON", NDefines.NCallsToAction.CALL_TO_ACTION_DO_NOTHING)

end


function wasTrueKeyEverInstalledOnTheDevice()
    tracer:Trace(TAG, "was truekey installed")
    -- Does true key already exist on the device?
    local thirdPartyAppHelper = appContext:getObject("Object.ThirdPartyAppHelper")
    if (thirdPartyAppHelper:isAppInstalled("com.mcafee.safeconnect.android")) then return true end

    -- If user installed true key atleast once on device do not show trigger
    if (thirdPartyAppHelper:wasAppInstalledPreviously("com.mcafee.safeconnect.android")) then return true end

    return false
end


function isAppSupportedByTrueKey(packageName)
    tracer:Trace(TAG, "isAppSupportedByTrueKey")
    local tkTeaserHelper = appContext:getObject("Object.TrueKeyTeaserHelper")
    if ( tkTeaserHelper:isAppSupportedByTrueKey(packageName) ) then return true end
    return false
end


function hasTrueKeySupportedAppRecentlyBeenOpened()
    tracer:Trace(TAG, "hasTrueKeySupportedAppRecentlyBeenOpened")
    -- Create local helper objects
    local appMonitorHelper = appContext:getObject("Object.AppMonitorHelper")
    -- Solution A: Check the last opened app only
   -- if ( isAppSupportedByTrueKey( appMonitorHelper:getLastOpenedApp() ) ) then return true end

    -- Solution B: Check the last few opened apps with a time window of 60 seconds
    local lastOpenedAppsList = appMonitorHelper:getLastOpenedApps()
    clearAppsList()
    for i = 0, ( lastOpenedAppsList:size() - 1 ) do
        if ( isAppSupportedByTrueKey( lastOpenedAppsList:get(i) ) ) then return true end
    end
    tracer:Trace(TAG, "hasTrueKeySupportedAppRecentlyBeenOpened returning false")
    return false
end


function isURLSupportedByTrueKey(urlName)
    tracer:Trace(TAG, "isURLSupportedByTrueKey")
    local tkTeaserHelper = appContext:getObject("Object.TrueKeyTeaserHelper")
    if ( tkTeaserHelper:isURLSupportedByTrueKey(urlName) ) then return true end
    tracer:Trace(TAG, "isURLSupportedByTrueKey false")
    return false
end


function hasTrueKeySupportedURLRecentlyBeenNavigated()

    -- Create local helper objects
    local urlMonitorHelper = appContext:getObject("Object.URLMonitorHelper")

    -- Solution A: Check the last navigated url only
   -- if ( isURLSupportedByTrueKey( urlMonitorHelper:getLastNavigatedURL() ) ) then return true end

    -- Solution B: Check the last few navigated urls with a time window of 60 seconds
    local lastNavigatedURLsList = urlMonitorHelper:getLastNavigatedUrls()
    clearUrlList()
    for i = 0, ( lastNavigatedURLsList:size() - 1 ) do
        if ( isURLSupportedByTrueKey( lastNavigatedURLsList:get(i) ) ) then return true end
    end

    return false
end


function updateTrueKeyMessagingStateForAppLaunch()
    -- Optimization
    if ( not canPromoteTrueKeyOnHomeScreen_intransigentChecks() ) then return end

    msgContext:addItemToList("MESSAGE_PROMOTE_TRUE_KEY_ON_HOME_SCREEN", "TK_SUPPORTED_APPS_LAUNCHED", event:getPackageName())
end


function updateTrueKeyMessagingStateForURLNavigate()
    tracer:Trace(TAG, "updateTrueKeyMessagingStateForURLNavigate")
    -- Optimization
    if ( not canPromoteTrueKeyOnHomeScreen_intransigentChecks() ) then return end
    tracer:Trace(TAG, "canPromoteTrueKeyOnHomeScreen_intransigentChecks")
    tracer:Trace(TAG, "addItemToList TK_SUPPORTED_URLS_NAVIGATED start ")
    msgContext:addItemToList("MESSAGE_PROMOTE_TRUE_KEY_ON_HOME_SCREEN", "TK_SUPPORTED_URLS_NAVIGATED", event:getUrl())
    tracer:Trace(TAG, "addItemToList TK_SUPPORTED_URLS_NAVIGATED end ")
end


function hasSufficientUserVisitsToTrueKeyCandidates()
    tracer:Trace(TAG, "hasSufficientUserVisitsToTrueKeyCandidates")
    tracer:Trace(TAG, "hasSufficientUserVisitsToTrueKeyCandidates")
    -- Doesn't matter for the first promotion
   if( not (msgContext:getMessageSeenCount("MESSAGE_PROMOTE_TRUE_KEY_ON_HOME_SCREEN") > 0 ) ) then return true end
    tracer:Trace(TAG, "hasSufficientUserVisit true")
    -- Matters for all subsequent promotions
    local tkCompatibleAppsVisited = msgContext:getList("MESSAGE_PROMOTE_TRUE_KEY_ON_HOME_SCREEN", "TK_SUPPORTED_APPS_LAUNCHED")
    local tkCompatibleUrlsVisited = msgContext:getList("MESSAGE_PROMOTE_TRUE_KEY_ON_HOME_SCREEN", "TK_SUPPORTED_URLS_NAVIGATED")
    local appSize  = tkCompatibleAppsVisited:size();
    local urlSize = tkCompatibleUrlsVisited:size();
    tracer:Trace(TAG, "appSize " .. appSize);
    tracer:Trace(TAG, "urlSize " .. urlSize);
    if( ( appSize + urlSize ) < NDefines.NCadence.MIN_COUNT_TRUEKEY_TARGETS ) then return false end
    tracer:Trace(TAG, "hasSufficientUserVisitsToTrueKeyCandidates true")
    return true

end

function clearAppsList()
    local appMonitorHelper = appContext:getObject("Object.AppMonitorHelper")
    appMonitorHelper:clearAppsList()
end

function clearUrlList()
    local urlMonitorHelper = appContext:getObject("Object.URLMonitorHelper")
    urlMonitorHelper:clearUrlList()
end

function handleOnTrueKeyPromoShownOnHomeScreen()
    -- Clear the historical state that no longer matters after this message
    msgContext:clearList("MESSAGE_PROMOTE_TRUE_KEY_ON_HOME_SCREEN", "TK_SUPPORTED_APPS_LAUNCHED")
    msgContext:clearList("MESSAGE_PROMOTE_TRUE_KEY_ON_HOME_SCREEN", "TK_SUPPORTED_URLS_NAVIGATED")

end